/*
 * Copyright (c) 2016 Couchbase, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * Copyright 2012 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package com.couchbase.client.core.logging;

import com.couchbase.client.deps.io.netty.util.internal.StringUtil;

import java.io.ObjectStreamException;
import java.io.Serializable;

import static com.couchbase.client.core.lang.backport.java.util.Objects.requireNonNull;

/**
 * A skeletal implementation of {@link CouchbaseLogger}.  This class implements
 * all methods that have a {@link CouchbaseLogLevel} parameter by default to call
 * specific logger methods such as {@link #info(String)} or {@link #isInfoEnabled()}.
 */
public abstract class AbstractCouchbaseLogger implements CouchbaseLogger, Serializable {

    private static final long serialVersionUID = -6382972526573193470L;

    private static final String EXCEPTION_MESSAGE = "Unexpected exception:";

    private final String name;

    private final RedactionLevel redactionLevel;

    /**
     * Creates a new instance.
     */
    protected AbstractCouchbaseLogger(String name, RedactionLevel redactionLevel) {
        this.name = requireNonNull(name, "name");
        this.redactionLevel = requireNonNull(redactionLevel, "redactionLevel");
    }

    @Override
    public String name() {
        return name;
    }

    @Override
    public boolean isEnabled(CouchbaseLogLevel level) {
        switch (level) {
            case TRACE:
                return isTraceEnabled();
            case DEBUG:
                return isDebugEnabled();
            case INFO:
                return isInfoEnabled();
            case WARN:
                return isWarnEnabled();
            case ERROR:
                return isErrorEnabled();
            default:
                throw new Error();
        }
    }

    @Override
    public void trace(Throwable t) {
        trace(EXCEPTION_MESSAGE, t);
    }

    @Override
    public void debug(Throwable t) {
        debug(EXCEPTION_MESSAGE, t);
    }

    @Override
    public void info(Throwable t) {
        info(EXCEPTION_MESSAGE, t);
    }

    @Override
    public void warn(Throwable t) {
        warn(EXCEPTION_MESSAGE, t);
    }

    @Override
    public void error(Throwable t) {
        error(EXCEPTION_MESSAGE, t);
    }

    @Override
    public void log(CouchbaseLogLevel level, String msg, Throwable cause) {
        switch (level) {
            case TRACE:
                trace(msg, cause);
                break;
            case DEBUG:
                debug(msg, cause);
                break;
            case INFO:
                info(msg, cause);
                break;
            case WARN:
                warn(msg, cause);
                break;
            case ERROR:
                error(msg, cause);
                break;
            default:
                throw new Error();
        }
    }

    @Override
    public void log(CouchbaseLogLevel level, Throwable cause) {
        switch (level) {
            case TRACE:
                trace(cause);
                break;
            case DEBUG:
                debug(cause);
                break;
            case INFO:
                info(cause);
                break;
            case WARN:
                warn(cause);
                break;
            case ERROR:
                error(cause);
                break;
            default:
                throw new Error();
        }
    }

    @Override
    public void log(CouchbaseLogLevel level, String msg) {
        switch (level) {
            case TRACE:
                trace(msg);
                break;
            case DEBUG:
                debug(msg);
                break;
            case INFO:
                info(msg);
                break;
            case WARN:
                warn(msg);
                break;
            case ERROR:
                error(msg);
                break;
            default:
                throw new Error();
        }
    }

    @Override
    public void log(CouchbaseLogLevel level, String format, Object arg) {
        switch (level) {
            case TRACE:
                trace(format, arg);
                break;
            case DEBUG:
                debug(format, arg);
                break;
            case INFO:
                info(format, arg);
                break;
            case WARN:
                warn(format, arg);
                break;
            case ERROR:
                error(format, arg);
                break;
            default:
                throw new Error();
        }
    }

    @Override
    public void log(CouchbaseLogLevel level, String format, Object argA, Object argB) {
        switch (level) {
            case TRACE:
                trace(format, argA, argB);
                break;
            case DEBUG:
                debug(format, argA, argB);
                break;
            case INFO:
                info(format, argA, argB);
                break;
            case WARN:
                warn(format, argA, argB);
                break;
            case ERROR:
                error(format, argA, argB);
                break;
            default:
                throw new Error();
        }
    }

    @Override
    public void log(CouchbaseLogLevel level, String format, Object... arguments) {
        switch (level) {
            case TRACE:
                trace(format, arguments);
                break;
            case DEBUG:
                debug(format, arguments);
                break;
            case INFO:
                info(format, arguments);
                break;
            case WARN:
                warn(format, arguments);
                break;
            case ERROR:
                error(format, arguments);
                break;
            default:
                throw new Error();
        }
    }

    protected Object readResolve() throws ObjectStreamException {
        return CouchbaseLoggerFactory.getInstance(name());
    }

    /**
     * Helper method to perform potential log redaction.
     *
     * If this method returns true, it did perform the redaction and logging. If
     * it returns false no side effect (no log line) has been sent to the logging
     * implementation.
     *
     * @param format the log string
     * @param arguments the log arguments
     * @return true if redacted, false otherwise.
     */
    protected boolean infoRedacted(final String format, final Object... arguments) {
        if (redactionLevel == RedactionLevel.NONE) {
            return false;
        }

        Object[] redacted = new Object[arguments.length];
        if (redactArgs(arguments, redacted)) {
            info(format, redacted);
            debug(format, arguments);
            return true;
        }

        return false;
    }

    /**
     * Helper method to perform potential log redaction.
     *
     * If this method returns true, it did perform the redaction and logging. If
     * it returns false no side effect (no log line) has been sent to the logging
     * implementation.
     *
     * @param format the log string
     * @param arguments the log arguments
     * @return true if redacted, false otherwise.
     */
    protected boolean warnRedacted(final String format, final Object... arguments) {
        if (redactionLevel == RedactionLevel.NONE) {
            return false;
        }

        Object[] redacted = new Object[arguments.length];
        if (redactArgs(arguments, redacted)) {
            warn(format, redacted);
            debug(format, arguments);
            return true;
        }

        return false;
    }

    /**
     * Helper method to perform potential log redaction.
     *
     * If this method returns true, it did perform the redaction and logging. If
     * it returns false no side effect (no log line) has been sent to the logging
     * implementation.
     *
     * @param format the log string
     * @param arguments the log arguments
     * @return true if redacted, false otherwise.
     */
    protected boolean errorRedacted(final String format, final Object... arguments) {
        if (redactionLevel == RedactionLevel.NONE) {
            return false;
        }

        Object[] redacted = new Object[arguments.length];
        if (redactArgs(arguments, redacted)) {
            error(format, redacted);
            debug(format, arguments);
            return true;
        }

        return false;
    }

    /**
     * Redact arguments for logging based on the policy level configured.
     *
     * @param arguments the input arguments.
     * @param redacted the redacted arguments injected, should be empty.
     * @return true if redacted at least one, false otherwise.
     */
    private boolean redactArgs(Object[] arguments, Object[] redacted) {
        boolean atLeastOneRedacted = false;
        for (int i = 0; i < arguments.length; i++) {
            Object arg = arguments[i];
            if (arg instanceof RedactableArgument) {
                arguments[i] = arg.toString();

                if (redactionLevel == RedactionLevel.FULL ||
                        (redactionLevel == RedactionLevel.PARTIAL
                                && ((RedactableArgument) arg).type() == RedactableArgument.ArgumentType.USER)
                        ) {
                    redacted[i] = "-REDACTED-";
                    atLeastOneRedacted = true;
                } else {
                    redacted[i] = arg.toString();
                }
            }
        }
        return atLeastOneRedacted;
    }



    @Override
    public String toString() {
        return StringUtil.simpleClassName(this) + '(' + name() + ')';
    }
}
